
/********************************************************************************
 *										*
 *				 M A C E K					*
 *		("Matroids Also Computed Efficiently" Kit)			*
 *										*
 * A set of tools and routines for computations with representable matroids.	*
 * Copyright (C) 2001,2002  Petr Hlineny.					*
 *										*
 *   This program is free software; you can redistribute it and/or modify	*
 *   it under the terms of the GNU General Public License as published by	*
 *   the Free Software Foundation; either version 2 of the License, or		*
 *   (at your option) any later version.					*
 *   You should have received a copy of the GNU General Public License		*
 *   along with this program; if not, write to the Free Software		*
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.	*
 * 										*
 * See the top-level "ABOUT" and "doc/ *" files for more information.		*
 *										*
 ********************************************************************************/














/**
 * Some theory about generating matroids for start...
 * The master description of the generating process is in ../include/gener.h.
 * 
 * This piece of code implements generating one-line extensions of the given matrix.
 * (A job that may seem simple, but it is not so simple in partial fields with infinitely
 * many elements -- then the choice for extension elements gets quite tricky...)
 * All possible extensions are generated; in the partial case even many
 * non-properly represented ones.
 * One may generate only a random sublist, or require representability testing for all
 * matrices.
 * 
 * See gener.c for additional (canonical) filtering applied to the generated extensions.
 * 
**/


#include "macek.h"
#include "gen.h"













/******************	Generating matrix extensions	*******************/
/**************************************************************************/


#define	CURDLEV		6		/*7 (use <=DEBUGLEV+2 to perform extra checks) */



/**
 * This function generates column/row (by tr=1/0) extensions to the matrix ex.
 * All distinct extension lines are generated here, normalized to unit-leading form
 * (unit for each component of a disconnected matrix).
 * When extending over a partial field, then also some pfield-checks of the matrix
 * are used to restrict the number of generated choices.
 * (But the extension is not guarranteed to be properly pfield-represented unless tst>=1!)
 * 
 * The generated extensions are collected in the list *els (if given).
 * (Remember to free the list and the matrices later.)
 * The function returns the number of generated extensions, or -1 on an error.
 * If ch>1 (>2), then the generated extensions are all checked and printed out.
 * 
 * If rnd>0, then only a random sublist of extensions is generated, about rnd
 * choices per each entry.
 * If tst==0, then all generated extensions are tested in a partial field randomly,
 * and if tst==1, then they are fully tested (all aplies only to partial fields).
**/

int	gener_matextens_ext(int ch, ematrix *ex, int tr, int rnd, int tst, ematrix ***els) {
	long		**ppat,**mpat;
	char		pprf[50];
	int		i,j,k,l, r,lc, ro,co,cco, rnf;
	int		*rcon,*rcpi,**radj, *selsg,*selex;
	ematrix		*ee, *edet, **rre, *rr;
	
	if (els?*els:0)  {PROGERROR("the list els must be given empty (null)"); return -1;}
	if (ch>1 && tst<1)  tst = 1;	/* (must test all extensions when printing output) */
	if (ch>2)  sprintf(pprf,"%.30s \t",printoutpref);
	if (ch>1)  OUTPUT("Going to generate %s extensions of the matrix %p[%s] in %s.\n",
			tr?"col":"row",ex,EMNAME(ex)?EMNAME(ex):"",pfield_curname());
	if (ch>1)  EMATOUTPUT(ex,printoutpref);
#ifndef FASTPROG
	DEBUG(CURDLEV-0+(ch<0),"Going to generate %s extensions of the matrix %p[%s] %dx%d in %s...\n",
			tr?"col":"row",ex,EMNAME(ex)?EMNAME(ex):"",ROWSM(ex),COLSM(ex),pfield_curname());
	if (rnd>0)  DEBUG(CURDLEV-0+(ch<0),"  - generating only random (%d) subset of extensions\n",rnd);
	if (tst>=0 && pfispartial)  DEBUG(CURDLEV-0+(ch<0),"  - %s testing extensions for proper representation\n",tst>0?"full":"");
	EMATDEBUGS(CURDLEV+1,ex,"  ex\t");
#endif
	if (!tr)  ematrix_transpose(ex);
	r = 0;
	j = (rnd>0? rnd: pfield_numfundamentals())+1;
	j = (pfispartial? j/(pfnumexp+1): j);
	for (i=0,k=1; i<ROWSM(ex); i++)
		if (k<GEN_MAXMATEXTENS)  k *= (j<2? 2:j);
	if (k>GEN_MAXMATEXTENS)  {PROGERROR("there are too many expected extensions to generate"); r = -11;}
	rnf = (rnd>0? (pfield_numfundamentals()+1)/rnd: 0);
	
		/**
		 * We prepare here working matrices ee (for the extended column) and
		 * edet (for choosing pfield value according to previous nonzero subdeterminants).
		 * Then we get the "connected" order rcon[] of rows of ex, component
		 * indexing in rcpi[], and the adjacency matrix of rows in radj[][].
		**/
	ro = ROWSM(ex);  cco = ro*ro;  co = COLSM(ex);
	ee = ematrix_copy_to(ex,ro,co+1);
	ee = ematrix_append_col(ee);
	edet = new_ematrix(ro,cco+1,ro,cco+1);
	rre = NULL;
	radj = pfispartial? malloc_twodim(sizeof(radj[0][0]),ro,ro):NULL;
	rcon = MMALLOC((2*ro+4)*sizeof(rcon[0]));
	rcpi = rcon+ro;
	j = struct_connorder_radj(ex,rcon,rcpi,radj);
	if (j<0)  DEBUG(CURDLEV-0+(ch<0),"  - generating extensions for a disconnected matrix\n");
	ppat = mpat = NULL;
		/**
		 * For a partial field, we prepare induced paths for all pairs of rows -
		 * the array ppat[][] contains the internal vertices of these paths,
		 * and mpat[][] the connecting columns for these paths.
		 * The rows/columns i are encoded by the bits (1l<<i).
		 * Additionally, we prepare refering matrices rre[] to ee which will
		 * be used for representability pre-checks in a partial field.
		**/
	if (pfispartial && r>=0) {
		ppat = malloc_twodim(sizeof(ppat[0][0]),ro,ro);
		mpat = malloc_twodim(sizeof(mpat[0][0]),ro,ro);
		j = struct_connpaths(ro,radj,rcon,ppat,mpat);
		if (j<0)  r = -2;
		else  j = gener_matextens_cdet(ex,rcon,ppat,mpat,edet);
		if (j<0)  r = -2;
		rre = MMALLOC(ro*sizeof(rre[0]));
		rr = ematrix_refer(ee,-1,-1,0,COLSM(ee));
		for (i=0; i<ro; i++) {
			ematrix_refadd_row(rr,rcon[i]);
			if (i<=0 || i>=ro-2 || GEN_USEEXTPREPF<=0)  rre[i] = NULL;
			else  rre[i] = ematrix_refer_all(rr);
		}
		dispose_ematrix(rr);
	}
	
		/**
		 * We construct the k-th entries of the added line for k=0,1,2,... here:
		 * We use functions pfield_get[pa]value() to guess possibilities
		 * for the entry inside gener_matextens_cycl();
		 * the arrays selsg[],selex[] keep our choices (-1 means the initial choice).
		 * Moreover, we check that the first nonzero entry in each conn component
		 * of rows of ee is always 1, other nonzeros are then ignored.
		 * The matrix edet is used here over and over as the working storage.
		 * The new entries are stored into ee later.
		 * Possible representability pre-check in a partial field is carried out
		 * later, and then we continue in the backtracking search.
		**/
	selsg = MMALLOC((2*ro+4)*sizeof(selsg[0]));
	selex = selsg+ro+2;
	k = 0;
	selsg[0] = selex[0] = -1;
	while (k>=0 && r>=0) {
		if (k>0?rcpi[k]<rcpi[k-1]:0)  {PROGERROR("The components of rows of ex must come in order!");}
		lc = 0;		/* (looking for the first nonzero in the component - must be 1) */
		for (j=k-1; j>=0 && SIGNM(edet,j,cco)==0; j--) ;
		if (j<0?1:rcpi[k]>rcpi[j])  lc = 1;
				/* actual selection of the next entry happens in this call: */
		l = gener_matextens_cycl(k,ppat,edet,selsg,selex,lc,rnf);
		if (l<=0) {	/* cycling (l=0) or going back in backtracking (l<0) */
			k += l;  continue;
		}
		COPYEXSIGM(ee,rcon[k],co,edet,k,cco);
		if (rre?rre[k]:0) {	/* representability pre-check for pfields - see rre[] above */
			if (!pfield_isequal(EXPM(edet,k,cco),SIGNM(edet,k,cco),EXPM(rre[k],k,co),SIGNM(rre[k],k,co)))
				{PROGERROR("Wrong rre[%d] matrix for repr pre-check!",k);}
			if (ematrix_inpfield_randlast(rre[k],1,1)<0) {
				DEBUG(CURDLEV+1+(ch<0),"  X repr pre-check rejected at k=%d\n",k);
				continue;
			}
		}
		if (k<ro-1) {		/* going forward in backtracking - new cycle at level k+1 */
			++k;  selsg[k] = selex[k] = -1;
			continue;
		}
			/**
			 * After we choose all entries of the extended column (in edet, and
			 * copied to ee above), we possibly apply the final representability
			 * test, and then we output or store the extension.
			 * The return value r counts the number of extensions found here.
			**/
		if (tst>=0 && pfispartial) {
			j = (tst>0? ematrix_inpfield_last(ee,0,1):ematrix_inpfield_randlast(ee,0,1));
			if (j<0)  continue;
		}
		//if (CURDLEV+1<=DEBUGLEV || ch==2)  sxt = gener_extprintline(0,ee,1);  else  sxt = "";
		DEBUG(CURDLEV+1+(ch<0)," > extension #%d found:  %s\n",r+1,gener_extprintline(-1,ee,1));
		rr = ematrix_copy(ee);
		if (!tr)  ematrix_transpose(rr);
		r++;
		if (ch>=2)  OUTPUT(" > Extension #%d to [%s] found:  %s\n",r,EMNAME(ex)?EMNAME(ex):"",gener_extprintline(1,ee,1));
		if (ch>2)  EMATOUTPUTS(rr,pprf);
		if (els)  *els = alist_append(*els,rr);
		else  dispose_ematrix(rr);
	}
	
		/**
		 * Finally, we perform some late tests for correctness, and clean
		 * all data and matrices used in the generating process above.
		 * However, we must first dispose all rre[] matrices since they refer to
		 * ee which is going to be modified here.
		**/
	if (rre) {
		for (i=0; i<ro; i++)  if (rre[i])  dispose_ematrix(rre[i]);
		FREE(rre);
	}
	ematrix_remove_col(ee,co);
	if (!ematrix_isequal(ee,ex))  {PROGERROR("The matrix ee was damaged here!");}
	if (!tr)  ematrix_transpose(ex);
	if (ch>0)  OUTPUT("Generated %s %d %s-extensions of the matrix %p [%s] (%dx%d).\n",
			rnd>0?"random":"total",r,pfield_curname(),ex,EMNAME(ex)?EMNAME(ex):"",ROWSM(ex),COLSM(ex));
	if (ch>2)  OUTPUT("\n");
#ifndef FASTPROG
	DEBUG(CURDLEV-1+(ch<0),"... All done, generated %d %s-extensions of %p [%s] (%dx%d).\n",
			r,pfield_curname(),ex,EMNAME(ex)?EMNAME(ex):"",ROWSM(ex),COLSM(ex));
	if (ch>=0 && r>=0 && pfispartial && rnd<=0 && tst>=1) {
		rr = ematrix_copy(ex);  ematrix_swaprows(rr,0,ROWSM(rr)-1);
		if (ROWSM(rr)>3)  ematrix_swaprows(rr,1+RANDOM()%3,ROWSM(rr)-2);
		j = gener_matextens_ext(-1,rr,tr,0,1,NULL);
		dispose_ematrix(rr);
		if (j!=r)  {PROGERROR("Wrong number of extensions in a recursive test! %d!=%d",r,j);}
	}
	if (r>=0 && !pfispartial && rnd<=0 && rcpi[ro-1]==rcpi[0]) {
		j = pfield_numfundamentals()+1;
		for (i=0,k=1; i<ro; i++)  k *= j;
		k = (k-1)/(j-1)+1;
		if (r!=k)  {PROGERROR("Wrong total numer of extensions ?!? %d!=%d\n",r,k);}
	}
#endif
	if (r<0)  {PROGERROR("Generating matrix extensions has failed! %d",r);}
	dispose_ematrix(ee);		/* (used in generating extensions) */
	if (edet)  dispose_ematrix(edet);
	if (selsg)  FREE(selsg);
	if (ppat)  FREE(ppat);		/* (used to store "connected order" information) */
	if (mpat)  FREE(mpat);
	if (rcon)  FREE(rcon);
	if (radj)  FREE(radj);
	return r;
}



/**
 * Here we compute a "connecting subdeterminants" for generating extensions of the matrix ex.
 * This function is reasonable only for a (real) partial field.
 * We store the connecting subdeterminants in the matrix edet which was allocated before.
 * The rows of edet correspond to the rows of ex, but in the connected order given in rcon[].
 * Each column of edet corresponds to a pair of rows of ex.
 * 
 * The practical meaning of edet and ppat,mpat is the following:
 * If r1,r2 are two rows of edet, both entries l1,l2 of rows rcon[r1],rcon[r2] in the last column
 * in ex are nonzero, and other entries in the last column of ex corresponding to 1's in the
 * bitfield ppat[r1][r2] are zero, then the entries d1,d2 in
 * the column r1,r2 of edet at rows r1,r2 are nonzero and d1*l2-l1*d2 is equal to
 * a (+-) subdeterminat of the extended matrix ee.
 * (This allows to guess possible values of l2 knowing l1 and d1,d2 in a partial field.)
 * 
 * The return value is -1 for an error here, or 1 after normal computation.
**/

int	gener_matextens_cdet(ematrix *ex, int rcon[], long **ppat, long **mpat, ematrix *edet) {
	int		i,j, z,m,n, ro,co;
	long		zp,zm;
	ematrix		*re;
	exp_t		xx;
	sign_t		gg;
	
	ro = ROWSM(ex);  co = COLSM(ex);
	if (ro>=NUM_LONG_BITS-2 || co>=NUM_LONG_BITS-2)  {PROGERROR("The matrix is too large to fit into long bits!"); return -1;}
	DEBUG(CURDLEV+0," Computing \"connecting\" subdeterminants for guessing pfield values...\n");
		/**
		 * For each pair of rows i,j, we compute the "connecting" subdeterminants
		 * of ex which have columns given by the bits of mpat[i][j], and rows by the
		 * bits of ppat[i][j] plus the row j or i, resp.
		 * These subdeterminants are then used later to guess pfield entries in ex.
		 * The tricky part here is with signs of these subdeterminants since the rows
		 * in rcon[] are swapped -- so we compute the subdeterminants like if the
		 * end-rows of the path rcon[i],rcon[j] were after all internal rows.
		 * (Only relative sign is significat for us, so this is a correct approach.)
		**/
	for (i=0; i<ro; i++)  for (j=i+1; j<ro; j++) {
		zp = ppat[i][j];  zm = mpat[i][j];
		if (zp<0)  continue;
		for (z=0,m=i; z<2; z++,m=j) {
			re = ematrix_refer_empty(ex);
			for (n=0; n<ro; n++)  if ((zp&(1l<<n))!=0)
				ematrix_refadd_row(re,rcon[n]);
			ematrix_refadd_row(re,rcon[m]);	
			for (n=0; n<co; n++)  if ((zm&(1l<<n))!=0)
				ematrix_refadd_col(re,n);
			if (ROWSM(re)!=COLSM(re))  {PROGERROR("The subdeterminant must be square for an induced path!"); return -1;}
			ematrix_determinant(re,&xx,&gg);
			if (gg==0)  {PROGERROR("The subdeterminant cannot be zero for an induced path!"); return -1;}
			SETEXSIGM(edet,m,i+j*ro,xx,gg);
			SETEXSIGM(edet,m,j+i*ro,xx,gg);
			dispose_ematrix(re);
		}
	}
#ifndef FASTPROG
	DEBUG(CURDLEV+1,"  - relative 2x2determinant connections :\n");
	if (CURDLEV+1<=DEBUGLEV)  for (i=0; i<ro; i++)  for (j=i+1; j<ro; j++)
		SDEBUG(CURDLEV+1,"\t\t [#%d (%d), #%d (%d)]:\t ppat=%lX mpat=%lX,  det1=%s  det2=%s\n",
			i,ROWSID(ex,rcon[i]), j,ROWSID(ex,rcon[j]), ppat[i][j],mpat[i][j],
			pfield_printvalue(16,EXPM(edet,i,i+j*ro),SIGNM(edet,i,i+j*ro)),
			pfield_printvalue(16,EXPM(edet,j,i+j*ro),SIGNM(edet,j,i+j*ro)));
	DEBUG(CURDLEV+1,"Done with \"connecting\" subdeterminants.\n");
#endif
	return 1;
}


/**
 * This function implements one choice of an entry in gener_matextens_ext() above.
 * One call decides an entry at the row k of the extended column of edet.
 * The values selsg[k],selex[k] tell the last/next choice of a sign and exponent for the entry.
 * (Values selsg[k],selex[k]==-1 start choosing.)
 * The "connecting subdeterminants" (see gener_matextens_cdet()) edet and path patterns ppat[]
 * are used to guess allowable entries over a partial field.
 * Moreover, the function pre-checks representability over a partial field using other
 * suitable connecting paths.
 * The first nozero entry in each component of the rows is always set to 1 by call with sel1>0.
 * If rnf>0 is given, then only random subselection is considered, see below.
 * 
 * The return value -1 means to go back in backtracking (no more choices here),
 * 0 means to try a new iteration,
 * and 1 means a successful choice -- to go forward or record the whole extension.
**/

int	gener_matextens_cycl(int k, long **ppat, ematrix *edet,
					int selsg[], int selex[], int sel1, int rnf) {
	int	i,j,jf=0,r, ro,cco;
	long	zcp;
	exp_t	xx,xo;
	sign_t	gg,go;
	
	if (pfispartial && (!edet || !ppat))  {PROGERROREXIT("Must get edet when in partial field!");}
	ro = ROWSM(edet);  cco = COLSM(edet)-1;
	for (zcp=0,i=0; i<ro; i++)		/* (get zero pattern of the extended column) */
		if (i>=k || SIGNM(edet,i,cco)!=0)  zcp |= (1l<<i);
	
		/**
		 * The current values of selsg[k] and selex[k] tell what is our next choice:
		 * selsg[k]==1 chooses a nonzero entry indexed by selex[k] (1...max) in
		 * pfield_get[pa]value(), and selsg[k]==0 chooses a zero entry.
		 * The select values of -1 mean the initial choice for selsg[k] and selex[k].
		 * Here we (possibly) move to the next selection of the sign, and set the
		 * number of exponent choices.
		 * We also try a random jump on the exponent choices for rnf>0.
		**/
	if (selex[k]<=0)
		if (++selsg[k]>1) {	/* the next value of the sign (trying in order 0,+-1) */
			DEBUG(CURDLEV+3,"%*s calling \"cycled\" at k=%d:  ^back^\n",2*k,"",k);
			return -1;	/* (choices are exhausted...) */
		}
	DEBUG(CURDLEV+3,"%*s calling \"cycle\" at k=%d: selsg=%d, selex=%d, sel1=%d, zcp=%lX\n",2*k,"",k,selsg[k],selex[k],sel1,zcp);
		/**
		 * The meaning of selex[],selsg[] is explained above, and
		 * the meaning of edet and ppat,mpat for a partial field is explained in
		 * gener_matextens_cdet().
		 * Here we either choose a zero entry at k (selsg[k]==0), or we choose the
		 * selex[k]-th nonzero entry at k, possibly relative to the "connecting"
		 * subdeterminants stored in edet when in a partial field.
		 * In the latter case, we use the first induced path joining k with a previous
		 * nonzero across all internal zeroes.
		**/
	if (selsg[k]<=0) {
		selex[k] = 1;  r = 0;
		go = 0;  pfield_setzeroexp(&xo);
	} else if (sel1) {
		selex[k] = 1;  r = 0;
		go = 1;  pfield_setzeroexp(&xo);
	} else if (!pfispartial) {
		if (selex[k]<=0)  selex[k] = pfield_getavalue_num();
		if (rnf>0)  selex[k] -= RANDOM()%(rnf+1);
		r = (selex[k]<=0?-1: pfield_getavalue(selex[k],&xo,&go));
	} else {
		for (jf=0; jf<k; jf++)
			if ((zcp&(1l<<jf))!=0 && (zcp&ppat[jf][k])==0)  break;
		if (jf>=k)  {PROGERROR("No induced connecting path found for the next entry, why ?!?"); return -1;}
		pfield_mul(1,EXPM(edet,jf,cco),SIGNM(edet,jf,cco),
				1,EXPM(edet,k,jf+k*ro),SIGNM(edet,k,jf+k*ro),&xx,&gg);
		pfield_mul(-1,EXPM(edet,jf,jf+k*ro),-SIGNM(edet,jf,jf+k*ro),
				1,xx,gg,&xx,&gg);
		if (selex[k]<=0)  selex[k] = pfield_getpvalue_num(xx,gg);
		if (rnf>0) if (RANDOM()%(pfnumexp+1)==0)
			selex[k] -= RANDOM()%(rnf+1);
		r = (selex[k]<=0?-1: pfield_getpvalue(selex[k],xx,gg,&xo,&go));
	}
	--selex[k];			/* (the next choice for the exponent on the next call) */
	SETEXSIGM(edet,k,cco,xo,go);
	
		/**
		 * After choosing the extended entry, we try few more pfiel-oriented tests.
		 * We use the remaining (suitable) connecting determinants, for pfield-def checks.
		 * If an undefined subdeterminant is found here, then this extension is dead.
		**/
	if (go!=0 && !sel1 && pfispartial && edet && ppat) {
		for (j=0; j<k && r>=0; j++) {
			if ((zcp&(1l<<j))!=0 && (zcp&ppat[j][k])==0)
				r = ematrix_determinant_2x2check(edet,j,k,j+k*ro,cco);
			if (j<=jf && r<0)  {PROGERROR("Wrong setting of the new entry ?!?");}
		}		/* the first choice (at row jf found above) is supposed to be set OK */
	}
	if (r<0)  return 0;		/* (if the current choice at row k has failed, then try again) */
	DEBUG(CURDLEV+2,"%*s    -> chosen at #%d; l.%d : %s (selex %d)\n",2*k,"",k,k,pfield_pvalue(16,xo,go),selex[k]+1);
	return 1;		/* successfull extension is returned */
}



/**
 * This is a supplementary function for an easy printing of the extended line in the matrix e.
 * The last row is printed (as comma-separated numbers) if trq==0, the last column if trq==1.
 * The resulting string is returned in a static buffer - do not free(!!!) it.
**/

char*	gener_extprintline(int ch, ematrix *e, int trq) {
	int		i;
	static char	buf[200];
	
	if (!trq)  ematrix_transpose(e);
	buf[0] = 0;
	for (i=0; i<ROWSM(e) && strlen(buf)<160; i++) {
		if (ch<0)  pfield_pvalue_to(buf+strlen(buf),7,EXPM(e,i,COLSM(e)-1),SIGNM(e,i,COLSM(e)-1));
		else  pfield_printvalue_to(buf+strlen(buf),(ch>0?14:9),EXPM(e,i,COLSM(e)-1),SIGNM(e,i,COLSM(e)-1));
		if (i<ROWSM(e)-1)  strcat(buf,",");
	}
	if (i<ROWSM(e))  strcat(buf,"...");
	if (!trq)  ematrix_transpose(e);
	return buf;
}

































