
/********************************************************************************
 *										*
 *				 M A C E K					*
 *		("Matroids Also Computed Efficiently" Kit)			*
 *										*
 * A set of tools and routines for computations with representable matroids.	*
 * Copyright (C) 2001,2002  Petr Hlineny.					*
 *										*
 *   This program is free software; you can redistribute it and/or modify	*
 *   it under the terms of the GNU General Public License as published by	*
 *   the Free Software Foundation; either version 2 of the License, or		*
 *   (at your option) any later version.					*
 *   You should have received a copy of the GNU General Public License		*
 *   along with this program; if not, write to the Free Software		*
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.	*
 * 										*
 * See the top-level "ABOUT" and "doc/ *" files for more information.		*
 *										*
 ********************************************************************************/















/**
 * Some theory for start... about "magic values" for (elements of) matrices and matroids.
 * (Read general description of "structural" functions on the top of ../include/struct.h.)
 * 
 * When working with matrices or with the underlying matroids, it is often useful to
 * consider some "magic numbers" associated with these objects.
 * The magic numbers are, by default, invariant on the operations that we consider
 * (that means scaling and permuting lines for matrices, or taking various representations
 * for an abstract matroid in the second case).
 * We may then apply various tricks using symmetric functions to these numbers.
 * 
 * Here we provide several useful and reasonably fast functions for computing such
 * magic values for matrix lines, or for matroid elements and whole matroids.
 * These functions capture the structure of small zero subdeterminants, or of
 * small-rank flats in a matroid.
 * (A similar "matroid hash" function is implemented inside ematrix_printmore()
 * in ../ematrix/ematexp.c, but that hash function is exponentially slow to compute,
 * so it should not be used inside other computations here.)
 * 
**/






#include "macek.h"
#include "str.h"












/******************	Miscelanelous	*******************/
/**********************************************************/


#undef CURDLEV
#define CURDLEV		7		/* (need <=DEBUGLEV+2 to perform paranoic tests) */


/**
 * This function decides whether there is an injection from the array fr[] to the array to[]
 * such that elements of value other than nomag are mapped to elements of the same value.
 * The return value is 0 for no injection, or d+1 where d is the number of distinct values.
 * (We typically use nomag==0, ie. only nonzero values are considered, in both fr,to!)
**/

int	strmag_isinjection_ext(long *fr, int frn, long *to, int ton, long nomag) {
	int	i,j, p,m,r;
	long	*ar,arstack[30];
	
	if (!fr || !to || frn<0 || ton<0)  {PROGERROR("both magic arrays must be given here!");}
	if (frn>ton)  return 0;
	if (frn+ton<28)  ar = arstack;
	else  ar = MMALLOC((frn+ton+2)*2*sizeof(ar[0]));
	
	for (i=0; i<frn; i++)  ar[i] = fr[i];
	for (i=0; i<ton; i++)  ar[i+frn] = to[i];
	r = 1;
	for (i=0; i<frn && r>0; i++)  if (ar[i]!=nomag) {
		m = 0;  r++;  p = ar[i];
		for (j=i; j<frn+ton; j++)
			if (ar[j]==p) {
				(j<frn? m--: m++);
				ar[j] = nomag;
			}
		if (m<0)  r = 0;
	}
	if (ar && ar!=arstack)  FREE(ar);
	return r;
}


/**
 * Here we compute a simple symmetric function of the values in the array ar[] of length arn.
 * The result is always nonzero.
**/

long	strmag_symmfunction_ext(int lev, long *ar, int arn) {
	int	i;
	long	f1,f2,f3,f4, l;
	
	f1 = f2 = f3 = f4 = 0;
	for (i=0; i<arn; i++) {
		f1 += ar[i];  f2 += (l=ar[i]*ar[i]);
		if (lev>2)  f3 += l*ar[i];
		if (lev>3)  f3 += l*l;
	}
	l = f1%1001l + f2*1001l;
	if (lev>2)  l += f3*100001l;
	if (lev>3)  l += f4*10000001l;
	if (!l)  l = 1;
	return l;
}


/**
 * 
 * combining two- to one- hash values (straut.old l. 440)........
 * 
**/














/******************	Matrix values	*******************/
/**********************************************************/


#undef CURDLEV
#define CURDLEV		7		/* (need <=DEBUGLEV+2 to perform paranoic tests) */


/**
 * This function computes the zero-pattern of the lines (rows tr==0, columns tr==1) of
 * the matrix e; storing the zeros of each line into bits of an element of zp[],
 * and the number of zeros +1 in the line to zn[].
 * The total number of zeros +1 is returned from the function.
 * Only lines indexed s1...s2-1 are considered.
 * The arrays zn,zp must be allocated by the caller to length >=s2-s1.
 * If zn or zp is NULL, then these information are not stored.
**/

long	strmag_zerolines_ext(ematrix *e, int tr, int s1, int s2, long *zp, long *zn) {
	int	i,j;
	long	p,k,l;
	
	if (!tr)  ematrix_transpose(e);
	for (l=0,j=s1; j<s2; j++) {
		k = 0;  if (zp)  zp[j-s1] = 0l;
		for (i=0; i<ROWSM(e); i++) {
			if (SIGNM(e,i,j)==0)  k++;
			else if (zp)  zp[j-s1] ^= (1l<<(i%NUM_LONG_BITS));
		}
		l += k;  if (zn)  zn[j-s1] = k+1;
	}
	if (!tr)  ematrix_transpose(e);
	return l+1;
}


/**
 * Here we find the clusters of zero 2x2 subdeterminants of the two given rows r1,r2 of ee.
 * (The zero subdeterminants form cliques if viewed as edges of a graph.)
 * The same is computed for columns if tr==1.
 * The result is stored in ar[]; ar[i] contains the index of the first column j<i for
 * which the subdeterminant of columns i,j is zero, or ar[i]=i if there is no zero.
 * The columns of both-0 entries are treated specially - they form their own cluster.
 * Subdeterminants may be undefined, then they are treated as nonzeros.
 * 
 * The function returns the number of zero 2x2 subdeterminants found +1.
 * If ar==NULL is given, then only this number is returned.
 * (The implementation may look complicated here, but it is actually faster than the trivial method.)
**/

long	strmag_twolinedet_ext(ematrix *e, int tr, int r1, int r2, int *ar) {
	int	i,j,i0,k, p,q, *nr, nrstack[30];
	sign_t	gg;
	
	if (tr)  ematrix_transpose(e);
	if (COLSM(e)<12)  nr = nrstack;
	else  nr = MMALLOC((2*COLSM(e)+4)*sizeof(nr[0]));
	if (!ar)  ar = nr+COLSM(e);
	
	i0 = -1;	/* i0 keeps the first both-0 column, ar[0] starts with 0 automatically in the cycle */
	for (i=0; i<COLSM(e); i++) {
		if (SIGNM(e,r1,i)==0 && SIGNM(e,r2,i)==0) {
			ar[i] = i0 = (i0==-1? i:i0);
			continue;		/* (both-0 columns need special treatment!) */
		}
		q = (SIGNM(e,r1,i)!=0 && SIGNM(e,r2,i)!=0);
		for (j=0; j<i; j++)  if (ar[j]==j && ar[j]!=i0) {
			if (q!=(SIGNM(e,r1,j)!=0 && SIGNM(e,r2,j)!=0))
				continue;	/* (det cannot be 0 with exactly one 0 entry) */
			p = ematrix_determinant_2x2ch(0,e,r1,r2,j,i,NULL,&gg);
			if (p>=0 && gg==0)  break;
		}	/* the subdeterminant may not be defined here, but then treated as nonzero */
		ar[i] = j;
	}
	for (i=0; i<COLSM(e); i++)  nr[i] = 0;
	for (i=k=0; i<COLSM(e); i++)  k += nr[ar[i]]++;
	if (i0>=0)  k += nr[i0]*(COLSM(e)-nr[i0]);
	
#ifndef FASTPROG
	if (IFRANDDEBUGLESS(222)) {
	  for (i=q=0; i<COLSM(e); i++) for (j=0; j<i; j++) {
		p = ematrix_determinant_2x2ch(0,e,r1,r2,j,i,NULL,&gg);
		q += (p>=0 && gg==0);
	  }if (q!=k)  {PROGERROR("something wrong with twolinedet -- number of zeros %d!=%d",k,q);}
	}
	if (CURDLEV+3<=DEBUGLEV) {
		SDEBUG(CURDLEV,"\t\t\ttwolinedet %d (+1) at %d,%d: ar[] = ",k+1,r1,r2);
		for (i=0; i<COLSM(e); i++)  SDEBUG(CURDLEV," %d,",ar[i]); SDEBUG(CURDLEV,"\n"); }
#endif
	if (nr && nr!=nrstack)  FREE(nr);
	if (tr)  ematrix_transpose(e);
	return k+1;
}


/**
 * This is a magic number computed from all entries of the given matrix.
 * It is invariant only on entry-wise equality.
**/

long	strmag_matrixentries_ext(ematrix *e) {
	long	ar[40], l=38;
	int	i,j;
	
	for (i=0; i<l; i++)  ar[i] = 0l;
	for (i=0; i<ROWSM(e); i++)  for (j=0; j<COLSM(e); j++)
		ar[(4*i+j)%l] += (long)pfield_hashvalue(EXPM(e,i,j),SIGNM(e,i,j));
	return strmag_symmfunction(ar,l);
}



/**
 * 
 * 
**/















/******************	Abstract matroid values	*******************/
/******************************************************************/


#undef CURDLEV
#define CURDLEV		7		/* (need <=DEBUGLEV+2 to perform paranoic tests) */


/**
 * This function computes magic numbers for matroid elements based on small flats in
 * the matroid.
 * These numbers are matroid-invariants!
 * The numbers are returned in the array flt[] (indexed first by rows, then by columns).
 * 
 * This function tries (co)closures of all k-tuples of elements (k = minr...maxr-1),
 * and records numbers based on the sizes of nontrivial ones.
**/

long	strmag_flatlines_ext(ematrix *e, int minr, int maxr, int tf, long flt[]) {
	ematrix		*ec, **el,**x;
	long		r,a;
	int		i,j, k,l, p,t;
	
	if (maxr<=minr || maxr>minr+5)  maxr = minr+2;
	if (flt)  for (i=0; i<ROWSM(e)+COLSM(e); i++)  flt[i] = 0;
	DEBUG(CURDLEV+2,"Computing flatlines for the matroid %p [%s] (%dx%d)...\n",e,EMNAME(e)?EMNAME(e):"",ROWSM(e),COLSM(e));
	r = 0;
	for (k=minr, p=0; k<maxr && p<tf; k++) {
		el = ematrix_submatrices_sub(e,k);
				/**
				 * Here we cycle all k-element submatrices of e, look at nontrivial
				 * closures and coclosures, and store a hash based on their
				 * sizes to all closure elements.
				**/
		for (x=el; x?*x:0; x++)  for (t=0; t<2; t++) {
			ec = ematrix_closure_tr(e,*x,t);
			i = ROWSM(ec)+COLSM(ec);
			if (i>k)  for (p++, j=0; j<i; j++) {
				a = (i*i+t*i+1l)<<(4*(k-minr));
				r += a;
				l = (j<ROWSM(ec)? GETREFMROW(ec,j): GETREFMCOL(ec,j-ROWSM(ec))+ROWSM(e));
				if (flt)  flt[l] += a;
			}
			dispose_ematrix(ec);
		}
		dispose_alist_mats(el);
	}
#ifndef FASTPROG
	if (CURDLEV+2<=DEBUGLEV && flt) {
		SDEBUG(CURDLEV+2,"\t\t(fll up to k=%d, p=%d) flt[] =\t",k-1,p);
		for (i=0; i<ROWSM(e)+COLSM(e); i++)  SDEBUG(CURDLEV+2,"%ld, ",flt[i]);
		SDEBUG(CURDLEV+2,"\n"); }
#endif
	return r;
}




/**
 * 
 * 
**/



/**
 * 
 * 
**/







































